import sys

sys.path.append(r'/Users/csvega/Desktop/Madison/Code/Simulations')
sys.path.append(r'/Users/csvega/Desktop/Madison/Code/Vlasov-Maxwell_Spectral_Solver/Vlasov-MaxwellSpectralSolver')

import json
import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from JAX_VM_solver import VM_simulation
from initialize_in_xv_space import initialize_system_xv
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
import time

   
# Simulation parameters.
Nx, Ny, Nz = 33, 33, 1
Lx, Ly, Lz = 50.0, 50.0, 1.0
Nn, Nm, Np, Ns = 4, 4, 4, 2 
mi_me, Ti_Te = 25.0, 1.0
Omega_cs = jnp.array([0.5, 0.5 / mi_me])
qs = jnp.array([-1, 1])
alpha_e = jnp.array([0.25, 0.25, 0.25])
alpha_s = jnp.concatenate([alpha_e, alpha_e * jnp.sqrt(Ti_Te / mi_me)])
u_s = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
nu = 1.0
t_steps, t_max, dt = 5, 0.4, 0.05



# Initial conditions

vte = alpha_s[0] / jnp.sqrt(2) # Electron thermal velocity.
vti = vte * jnp.sqrt(1 / mi_me) # Ion thermal velocity.
deltaB = 0.2 # In-plane magnetic field amplitude. 
U0 = deltaB * Omega_cs[0] / jnp.sqrt(mi_me) # Fluid velocity amplitude.

# Wavenumbers.
kx = 2 * jnp.pi / Lx
ky = 2 * jnp.pi / Ly

# Electron and ion fluid velocities.
Ue = lambda x, y, z: U0 * jnp.array([-jnp.sin(ky * y), jnp.sin(kx * x), -deltaB * Omega_cs[0] * (2 * kx * jnp.cos(2 * kx * x) + ky * jnp.cos(ky * y))])
Ui = lambda x, y, z: U0 * jnp.array([-jnp.sin(ky * y), jnp.sin(kx * x), jnp.zeros_like(x)])

# Magnetic and electric fields.
B = lambda x, y, z: jnp.array([-deltaB * jnp.sin(ky * y), deltaB * jnp.sin(2 * kx * x), jnp.ones_like(x)])
E = lambda x, y, z: jnp.array([jnp.zeros_like(x), jnp.zeros_like(x), jnp.zeros_like(x)]) # Is this consistent with fe, fi?

# Electron and ion distribution functions.
fe = (lambda x, y, z, vx, vy, vz: (1 / (((2 * jnp.pi) ** (3 / 2)) * vte ** 3) * 
                                jnp.exp(-((vx - Ue(x, y, z)[0])**2 + (vy - Ue(x, y, z)[1])**2 + (vz - Ue(x, y, z)[2])**2) / (2 * vte ** 2))))
fi = (lambda x, y, z, vx, vy, vz: (1 / (((2 * jnp.pi) ** (3 / 2)) * vti ** 3) * 
                                jnp.exp(-((vx - Ui(x, y, z)[0])**2 + (vy - Ui(x, y, z)[1])**2 + (vz - Ui(x, y, z)[2])**2) / (2 * vti ** 2))))

# # Save parameters into txt.
# with open('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\S13_Two_stream_instability_1D_HF.txt', 'w') as file:
#     file.write(f"Nx, Ny, Nz: {Nx}, {Ny}, {Nz}\n")
#     file.write(f"Nvx, Nvy, Nvz: {Nvx}, {Nvy}, {Nvz}\n")
#     file.write(f"Lx, Ly, Lz: {Lx}, {Ly}, {Lz}\n")
#     file.write(f"Nn, Nm, Np, Ns: {Nn}, {Nm}, {Np}, {Ns}\n")
#     file.write(f"mi_me: {mi_me}\n")
#     file.write(f"Omega_cs: {Omega_cs.tolist()}\n")
#     file.write(f"qs: {qs.tolist()}\n")
#     file.write(f"alpha_s: {alpha_s.tolist()}\n")
#     file.write(f"u_s: {u_s.tolist()}\n")
#     file.write(f"nu: {nu}\n")
#     file.write(f"t_steps, t_max: {t_steps}, {t_max}\n")
#     file.write(f"dt: {dt}\n")

Ck_0, Fk_0 = initialize_system_xv(B, E, fe, fi, Omega_cs[0], mi_me, alpha_s, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns)

# Run simulation.
start_time = time.time()
Ck, Fk, t = VM_simulation(Ck_0, Fk_0, qs, nu, Omega_cs, alpha_s, mi_me, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns, t_max, t_steps, dt)
end_time = time.time()
t.block_until_ready()
print(f"Runtime: {end_time - start_time} seconds")

# # Save results.
# jnp.save('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\Ck', Ck)
# jnp.save('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\Fk', Fk)
# jnp.save('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\\time', t)

####################################################################################################################################################
# Data analysis.



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'JAX-in-Cell'
copyright = '2025, UWPlasma'
author = 'UWPlasma'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# import sys
# from pathlib import Path
# sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
extensions = [
   'sphinx.ext.duration',
   'sphinx.ext.napoleon',
   'autoapi.extension'
]

autoapi_dirs = ['../spectrax']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'piccolo_theme'



import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.scipy.integrate import trapezoid



def compute_energy(Ck, Fk, Omega_ce, mi_me, alpha_s, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nvx, Nvy, Nvz, Nn, Nm, Np):
    
    plasma_energy_1 = (0.5 * alpha_s[0] * alpha_s[1] * alpha_s[2]) * ((0.5 * (alpha_s[0] ** 2 + alpha_s[1] ** 2 + alpha_s[2] ** 2) + 
                                                (u_s[0] ** 2 + u_s[1] ** 2 + u_s[2] ** 2)) * Ck[:, 0, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] + 
                                                jnp.sqrt(2) * (alpha_s[0] * u_s[0] * Ck[:, 1, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nn - 1) + 
                                                               alpha_s[1] * u_s[1] * Ck[:, Nn, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nm - 1) +
                                                               alpha_s[2] * u_s[2] * Ck[:, Nn * Nm, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Np - 1)) +
                                          (1 / jnp.sqrt(2)) * (alpha_s[0] ** 2 * Ck[:, 2, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nn - 1) * jnp.sign(Nn - 2) +
                                                               alpha_s[1] ** 2 * Ck[:, 2 * Nn, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nm - 1) * jnp.sign(Nm - 2) + 
                                                               alpha_s[2] ** 2 * Ck[:, 2 * Nn * Nm, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Np - 1) * jnp.sign(Np - 2)))
    
    plasma_energy_2 = (0.5 * mi_me *  alpha_s[3] * alpha_s[4] * alpha_s[5]) * ((0.5 * (alpha_s[3] ** 2 + alpha_s[4] ** 2 + alpha_s[5] ** 2) + 
                                                (u_s[3] ** 2 + u_s[4] ** 2 + u_s[5] ** 2)) * Ck[:, Nn * Nm * Np, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] + 
                                                jnp.sqrt(2) * (alpha_s[3] * u_s[3] * Ck[:, 1, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nn - 1) + 
                                                               alpha_s[4] * u_s[4] * Ck[:, Nn * Nm * Np + Nn, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nm - 1) +
                                                               alpha_s[5] * u_s[5] * Ck[:, Nn * Nm * Np + Nn * Nm, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Np - 1)) +
                                          (1 / jnp.sqrt(2)) * ((alpha_s[3] ** 2) * Ck[:, Nn * Nm * Np + 2, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nn - 1) * jnp.sign(Nn - 2) +
                                                               (alpha_s[4] ** 2) * Ck[:, Nn * Nm * Np + 2 * Nn, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Nm - 1) * jnp.sign(Nm - 2) + 
                                                               (alpha_s[5] ** 2) * Ck[:, Nn * Nm * Np + 2 * Nn * Nm, (Ny-1)//2, (Nx-1)//2, (Nz-1)//2] * jnp.sign(Np - 1) * jnp.sign(Np - 2)))
  
        
    plasma_energy = plasma_energy_1 + plasma_energy_2

    EM_energy = 0.5 * jnp.sum(jnp.abs(Fk) ** 2, axis=(-4, -3, -2, -1)) * Omega_ce[0] ** 2
    
    
    return plasma_energy, EM_energy


def compute_entropy(f1, f2, alpha_s, Lx, Ly, Lz, Nx, Ny, Nz, Nvx, Nvy, Nvz):

    x = jnp.linspace(0, Lx, Nx)
    y = jnp.linspace(0, Ly, Ny)
    z = jnp.linspace(0, Lz, Nz)
    vx = jnp.linspace(-4 * alpha_s[0], 4 * alpha_s[0], Nvx)
    vy = jnp.linspace(-4 * alpha_s[1], 4 * alpha_s[1], Nvy)
    vz = jnp.linspace(-4 * alpha_s[2], 4 * alpha_s[2], Nvz)
    
    entropy_1 = (trapezoid(trapezoid(trapezoid(trapezoid(trapezoid(trapezoid(
              (jnp.abs(f1) * jnp.log(jnp.abs(f1))), y, axis=-6) , x, axis=-5), z, axis=-4), vy, axis=-3), vx, axis=-2), vz, axis=-1))
    
    vx = jnp.linspace(-4 * alpha_s[3], 4 * alpha_s[3], Nvx)
    vy = jnp.linspace(-4 * alpha_s[4], 4 * alpha_s[4], Nvy)
    vz = jnp.linspace(-4 * alpha_s[5], 4 * alpha_s[5], Nvz)
    
    entropy_2 = (trapezoid(trapezoid(trapezoid(trapezoid(trapezoid(trapezoid(
              (jnp.abs(f2) * jnp.log(jnp.abs(f2))), y, axis=-6) , x, axis=-5), z, axis=-4), vy, axis=-3), vx, axis=-2), vz, axis=-1))
    

    return entropy_1 + entropy_2


import matplotlib.pyplot as plt
import jax.numpy as jnp
from ._inverse_transform import inverse_HF_transform
from matplotlib.animation import FuncAnimation

__all__ = ['plot']

def plot(output):
    time = output["time"]; k_norm = output["k_norm"]; Nm = int(output["Nm"])
    u_s = output["u_s"]; alpha_s = output["alpha_s"]; nu = output["nu"]
    Lx = output["Lx"]; Nx = output["Nx"]; Nn = int(output["Nn"]); 
    Np = int(output["Np"]); dn1 = output["dn1"]; Ck = output["Ck"]
    
    # Setup plots
    fig, axes = plt.subplots(2, 3, figsize=(15, 9))
    plt.subplots_adjust(hspace=0.2, wspace=0.2)
    fig.suptitle(rf'$kv_{{th,e}}/\omega_{{pe}} = {k_norm:.2},'
                 +rf'\nu = {nu}, u_e = {u_s[0]}, \alpha_e = {alpha_s[0]:.3},'
                 +rf'N_x = {Nx}, N_n = {Nn}, \delta n = {dn1}$', fontsize=14)
    
    # Energy plots
    axes[0, 0].plot(time, output["electric_field_energy"], label="Electric Field Energy")
    axes[0, 0].plot(time, output["kinetic_energy"], label="Kinetic Energy")
    axes[0, 0].plot(time, output["kinetic_energy_species1"], label="Kinetic Energy Species 1")
    axes[0, 0].plot(time, output["kinetic_energy_species2"], label="Kinetic Energy Species 2")
    axes[0, 0].plot(time, output["total_energy"], label="Total Energy")
    axes[0, 0].set(title="Energy", xlabel=r"Time ($\omega_{pe}^{-1}$)", ylabel="Energy (J)", yscale="log")#, ylim=[1e-5, None])
    axes[0, 0].legend()
    
    # Relative Energy Error
    axes[1, 0].plot(time[1:], jnp.abs(output["total_energy"][1:]-output["total_energy"][0])/(output["total_energy"][0]+1e-9), label="Relative energy error")
    axes[1, 0].set(xlabel=r"Time ($\omega_{pe}^{-1}$)", ylabel="Relative Energy Error", yscale="log")#, ylim=[1e-5, None])
    
    # Plot electron density fluctuation vs t.
    axes[1, 1].plot(time, (jnp.abs(output["dCk"][:, 0, 0, int((Nx-1) / 2) + 1, 0].imag)), label='$|\delta C^e_{000,k}|$', linestyle='-', linewidth=2.0)
    axes[1, 1].set(title='Species 1 density fluctuation', ylabel=r'$log(|\delta C^e_{000,k}|)$', xlabel=r'$t\omega_{pe}$', yscale="log")#, ylim=[1e-20, None])
    
    # Plot ion density fluctuation vs t.
    axes[1, 2].plot(time, (jnp.abs(output["dCk"][:, Nn, 0, int((Nx-1) / 2) + 1, 0].imag)), label='$|\delta C^i_{000,k}|$', linestyle='-', linewidth=2.0)
    axes[1, 2].set(title='Species 2 density fluctuation', ylabel=r'$log(|\delta C^i_{000,k}|)$', xlabel=r'$t\omega_{pe}$', yscale="log")#, ylim=[1e-20, None])
    
    # Electron Phase space plot
    i=0
    vx = jnp.linspace(-4 * alpha_s[0], 4 * alpha_s[0], 201)
    Vx, Vy, Vz = jnp.meshgrid(vx, jnp.array([0.]), jnp.array([0.]), indexing='xy')
    f1 = inverse_HF_transform(Ck[:, i*Nn:(i+1)*Nn, ...], Nn, Nm, Np, 
                                  (Vx - u_s[3*i]) / alpha_s[3*i], 
                                  (Vy - u_s[3*i+1]) / alpha_s[3*i+1], 
                                  (Vz - u_s[3*i+2]) / alpha_s[3*i+2])
    electron_phase_plot = axes[0, 1].imshow(jnp.transpose(f1[0, 0, :, 0, 0, :, 0]), extent=(0, Lx, vx[0], vx[-1]),
                                            cmap='jet', origin='lower', interpolation='sinc')
    plt.colorbar(electron_phase_plot, ax=axes[0, 1], label="$f_1$")
    axes[0, 1].set(xlabel="x/d_e", ylabel="v/c", title="Species 1 Phase Space")
    axes[0, 1].set_aspect('auto', adjustable='box')
    electron_phase_text = axes[0, 1].text(
        0.5, 0.9, "", transform=axes[0, 1].transAxes, ha="center", va="top",
        fontsize=12, bbox=dict(facecolor='white', alpha=0.7, edgecolor='none'))
    
    # Ion Phase space plot
    i=1
    vx = jnp.linspace(-4 * alpha_s[0], 4 * alpha_s[0], 201)
    Vx, Vy, Vz = jnp.meshgrid(vx, jnp.array([0.]), jnp.array([0.]), indexing='xy')
    f2 = inverse_HF_transform(Ck[:, i*Nn:(i+1)*Nn, ...], Nn, Nm, Np, 
                                  (Vx - u_s[3*i]) / alpha_s[3*i], 
                                  (Vy - u_s[3*i+1]) / alpha_s[3*i+1], 
                                  (Vz - u_s[3*i+2]) / alpha_s[3*i+2])
    ion_phase_plot = axes[0, 2].imshow(jnp.transpose(f2[0, 0, :, 0, 0, :, 0]), extent=(0, Lx, vx[0], vx[-1]),
                                            cmap='jet', origin='lower', interpolation='sinc')
    plt.colorbar(ion_phase_plot, ax=axes[0, 2], label="$f_2$")
    axes[0, 2].set(xlabel="x/d_e", ylabel="v/c", title="Species 2 Phase Space")
    axes[0, 2].set_aspect('auto', adjustable='box')
    ion_phase_text = axes[0, 2].text(
        0.5, 0.9, "", transform=axes[0, 2].transAxes, ha="center", va="top",
        fontsize=12, bbox=dict(facecolor='white', alpha=0.7, edgecolor='none'))
    
    def update(frame):
        electron_phase_plot.set_array(jnp.transpose(f1[frame, 0, :, 0, 0, :, 0]))
        electron_phase_plot.set_clim(vmin=f1[frame].min(), vmax=f1[frame].max())
        electron_phase_text.set_text(f"Time: {time[frame]:.1f} * ωₚ")
        
        ion_phase_plot.set_array(jnp.transpose(f2[frame, 0, :, 0, 0, :, 0]))
        ion_phase_plot.set_clim(vmin=f2[frame].min(), vmax=f2[frame].max())
        ion_phase_text.set_text(f"Time: {time[frame]:.1f} * ωₚ")
        
        return [electron_phase_plot, electron_phase_text, ion_phase_plot, ion_phase_text]
    
    ani = FuncAnimation(fig, update, frames=len(time), blit=True, interval=1, repeat_delay=1000)

    plt.show()



import jax.numpy as jnp

__all__ = ['diagnostics']

def diagnostics(output):
    alpha_s = output["alpha_s"]
    mi_me = output["mi_me"]
    u_s = output["u_s"]
    Fk = output["Fk"]
    Ck = output["Ck"]
    Lx = output["Lx"]
    Nx = output["Nx"]
    Ny = output["Nx"]
    Nz = output["Nx"]
    Omega_cs = output["Omega_cs"]
    Nn = output["Nn"]
    Nm = output["Nm"]
    Np = output["Np"]
    
    lambda_D = jnp.sqrt(1 / (2 * (1 / alpha_s[0] ** 2 + 1 / (mi_me * alpha_s[3] ** 2)))) # Debye length.
    k_norm = jnp.sqrt(2) * jnp.pi * alpha_s[0] / Lx # Perturbation wavenumber normalized to the inverse of the Debye length.
    
    half_nx = jnp.array((Nx-1)/2, int)
    half_ny = jnp.array((Ny-1)/2, int)
    half_nz = jnp.array((Nz-1)/2, int)

    kinetic_energy_species1 = (0.5 * alpha_s[0] * alpha_s[1] * alpha_s[2]) * ((0.5 * (alpha_s[0] ** 2 + alpha_s[1] ** 2 + alpha_s[2] ** 2) + 
                                                (u_s[0] ** 2 + u_s[1] ** 2 + u_s[2] ** 2)) * Ck[:, 0, half_ny, half_nx, half_nz] + 
                                                jnp.sqrt(2) * (alpha_s[0] * u_s[0] * Ck[:, 1, half_ny, half_nx, half_nz] * jnp.sign(Nn - 1) + 
                                                               alpha_s[1] * u_s[1] * Ck[:, Nn, half_ny, half_nx, half_nz] * jnp.sign(Nm - 1) +
                                                               alpha_s[2] * u_s[2] * Ck[:, Nn * Nm, half_ny, half_nx, half_nz] * jnp.sign(Np - 1)) +
                                          (1 / jnp.sqrt(2)) * (alpha_s[0] ** 2 * Ck[:, 2, half_ny, half_nx, half_nz] * jnp.sign(Nn - 1) * jnp.sign(Nn - 2) +
                                                               alpha_s[1] ** 2 * Ck[:, 2 * Nn, half_ny, half_nx, half_nz] * jnp.sign(Nm - 1) * jnp.sign(Nm - 2) + 
                                                               alpha_s[2] ** 2 * Ck[:, 2 * Nn * Nm, half_ny, half_nx, half_nz] * jnp.sign(Np - 1) * jnp.sign(Np - 2)))

    kinetic_energy_species2 = (0.5 * mi_me *  alpha_s[3] * alpha_s[4] * alpha_s[5]) * ((0.5 * (alpha_s[3] ** 2 + alpha_s[4] ** 2 + alpha_s[5] ** 2) + 
                                                (u_s[3] ** 2 + u_s[4] ** 2 + u_s[5] ** 2)) * Ck[:, Nn * Nm * Np, half_ny, half_nx, half_nz] + 
                                                jnp.sqrt(2) * (alpha_s[3] * u_s[3] * Ck[:, 1, half_ny, half_nx, half_nz] * jnp.sign(Nn - 1) + 
                                                               alpha_s[4] * u_s[4] * Ck[:, Nn * Nm * Np + Nn, half_ny, half_nx, half_nz] * jnp.sign(Nm - 1) +
                                                               alpha_s[5] * u_s[5] * Ck[:, Nn * Nm * Np + Nn * Nm, half_ny, half_nx, half_nz] * jnp.sign(Np - 1)) +
                                          (1 / jnp.sqrt(2)) * ((alpha_s[3] ** 2) * Ck[:, Nn * Nm * Np + 2, half_ny, half_nx, half_nz] * jnp.sign(Nn - 1) * jnp.sign(Nn - 2) +
                                                               (alpha_s[4] ** 2) * Ck[:, Nn * Nm * Np + 2 * Nn, half_ny, half_nx, half_nz] * jnp.sign(Nm - 1) * jnp.sign(Nm - 2) + 
                                                               (alpha_s[5] ** 2) * Ck[:, Nn * Nm * Np + 2 * Nn * Nm, half_ny, half_nx, half_nz] * jnp.sign(Np - 1) * jnp.sign(Np - 2)))
                                                    
    electric_field_energy = 0.5 * jnp.sum(jnp.abs(Fk) ** 2, axis=(-4, -3, -2, -1)) * Omega_cs[0] ** 2
    
    total_energy = kinetic_energy_species1 + kinetic_energy_species2 + electric_field_energy

    output.update({
        'lambda_D': lambda_D,
        'k_norm': k_norm,
        'kinetic_energy_species1': kinetic_energy_species1,
        'kinetic_energy_species2': kinetic_energy_species2,
        'kinetic_energy': kinetic_energy_species1 + kinetic_energy_species2,
        'electric_field_energy': electric_field_energy,
        'total_energy': total_energy,
    })
    


import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
number_of_processors_to_use = 5 # Parallelization, this should divide total resolution
os.environ["XLA_FLAGS"] = f'--xla_force_host_platform_device_count={number_of_processors_to_use}'
from time import time
import jax.numpy as jnp
from jax import block_until_ready
from spectrax import simulation, load_parameters, plot

# Read from input.toml
# input_parameters, solver_parameters = load_parameters('input_1D_landau_damping.toml')
toml_file = os.path.join(os.path.dirname(__file__), 'input_1D_landau_damping.toml')
input_parameters, solver_parameters = load_parameters(toml_file)

alpha_s = input_parameters["alpha_s"]
dn = input_parameters["dn1"]
Nx = solver_parameters["Nx"]
Nn = solver_parameters["Nn"]
kx = input_parameters["kx"]
Omega_cs = input_parameters["Omega_cs"]

# Fourier components of magnetic and electric fields.
Fk_0 = jnp.zeros((6, 1, Nx, 1), dtype=jnp.complex128)
Fk_0 = Fk_0.at[0, 0, int((Nx-1)/2-1), 0].set(dn / (2 * kx * Omega_cs[0]))
Fk_0 = Fk_0.at[0, 0, int((Nx-1)/2+1), 0].set(dn / (2 * kx * Omega_cs[0]))
input_parameters["Fk_0"] = Fk_0

# Hermite-Fourier components of electron and ion distribution functions.
Ce0_mk, Ce0_0, Ce0_k = 0 + 1j * (1 / (2 * alpha_s[0] ** 3)) * dn, 1 / (alpha_s[0] ** 3) + 0 * 1j, 0 - 1j * (1 / (2 * alpha_s[0] ** 3)) * dn
Ci0_0 = 1 / (alpha_s[3] ** 3) + 0 * 1j
Ck_0 = jnp.zeros((2 * Nn, 1, Nx, 1), dtype=jnp.complex128)
Ck_0 = Ck_0.at[0, 0, int((Nx-1)/2-1), 0].set(Ce0_mk)
Ck_0 = Ck_0.at[0, 0, int((Nx-1)/2), 0].set(Ce0_0)
Ck_0 = Ck_0.at[0, 0, int((Nx-1)/2+1), 0].set(Ce0_k)
Ck_0 = Ck_0.at[Nn, 0, int((Nx-1)/2), 0].set(Ci0_0)
input_parameters["Ck_0"] = Ck_0

# Simulate
start_time = time()
output = block_until_ready(simulation(input_parameters, **solver_parameters))
print(f"Runtime: {time() - start_time} seconds")

# Plot results
plot(output)

# print('Saving results...')
# jnp.savez('output_landau-damping.npz', **output)


import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
number_of_processors_to_use = 5 # Parallelization, this should divide total resolution
os.environ["XLA_FLAGS"] = f'--xla_force_host_platform_device_count={number_of_processors_to_use}'
from time import time
import jax.numpy as jnp
from jax import block_until_ready
from spectrax import simulation, load_parameters, plot

# Read from input.toml
# input_parameters, solver_parameters = load_parameters('input_1D_landau_damping.toml')
toml_file = os.path.join(os.path.dirname(__file__), 'input_1D_Weibel.toml')
input_parameters, solver_parameters = load_parameters(toml_file)

alpha_s = input_parameters["alpha_s"]
dEz = input_parameters["dEz"]
Nx = solver_parameters["Nx"]
Nn = solver_parameters["Nn"]
Nm = solver_parameters["Nm"]
Np = solver_parameters["Np"]
kx = input_parameters["kx"]
Omega_cs = input_parameters["Omega_cs"]

# Fourier components of magnetic and electric fields.
Fk_0 = jnp.zeros((6, 1, Nx, 1), dtype=jnp.complex128)
Fk_0 = Fk_0.at[2, 0, int((Nx-1)/2-1), 0].set(dEz / (2 * kx * Omega_cs[0]))
Fk_0 = Fk_0.at[2, 0, int((Nx-1)/2+1), 0].set(dEz / (2 * kx * Omega_cs[0]))
input_parameters["Fk_0"] = Fk_0

# Hermite-Fourier components of electron and ion distribution functions.
C10_0 = 1 / (alpha_s[0] ** 2 * alpha_s[2]) + 0 * 1j
C20_0 = 1 / (alpha_s[3] ** 2 * alpha_s[5]) + 0 * 1j
Ck_0 = jnp.zeros((2 * Nn * Nm * Np, 1, Nx, 1), dtype=jnp.complex128)
Ck_0 = Ck_0.at[0, 0, int((Nx-1)/2), 0].set(C10_0)
Ck_0 = Ck_0.at[Nn * Nm * Np, 0, int((Nx-1)/2), 0].set(C20_0)
input_parameters["Ck_0"] = Ck_0

# Simulate
start_time = time()
output = block_until_ready(simulation(input_parameters, **solver_parameters))
print(f"Runtime: {time() - start_time} seconds")

# Plot results
plot(output)

# print('Saving results...')
# jnp.savez('output_landau-damping.npz', **output)


import jax.numpy as jnp
from jax import jit, config, vmap
config.update("jax_enable_x64", True)
from jax.debug import print as jprint
from functools import partial
from diffrax import (diffeqsolve, Tsit5, Dopri5, ODETerm,
                     SaveAt, PIDController, TqdmProgressMeter)
from ._initialization import initialize_simulation_parameters
from ._model import plasma_current, Hermite_Fourier_system
from ._diagnostics import diagnostics

# Parallelize the simulation using JAX
from jax.sharding import PartitionSpec as P
from jax.experimental.shard_map import shard_map
from jax import devices, make_mesh, NamedSharding, device_put
mesh = make_mesh((len(devices()),), ("batch"))
spec = P("batch")
sharding = NamedSharding(mesh, spec)

__all__ = ["cross_product", "ode_system", "simulation"]

@jit
def cross_product(k_vec, F_vec):
    """
    Computes the cross product of two 3D vectors.
    Args:
        k_vec (array-like): A 3-element array representing the first vector (k).
        F_vec (array-like): A 3-element array representing the second vector (F).
    Returns:
        jnp.ndarray: A 3-element array representing the cross product k x F.
    """
    kx, ky, kz = k_vec
    Fx, Fy, Fz = F_vec
    return jnp.array([ky * Fz - kz * Fy, kz * Fx - kx * Fz, kx * Fy - ky * Fx])

@partial(jit, static_argnames=['Nx', 'Ny', 'Nz', 'Nn', 'Nm', 'Np', 'Ns'])
def ode_system(Nx, Ny, Nz, Nn, Nm, Np, Ns, t, Ck_Fk, args):
    (qs, nu, D, Omega_cs, alpha_s, u_s,
     Lx, Ly, Lz, kx_grid, ky_grid, kz_grid
    ) = args

    total_Ck_size = Nn * Nm * Np * Ns * Nx * Ny * Nz
    Ck = Ck_Fk[:total_Ck_size].reshape(Nn * Nm * Np * Ns, Ny, Nx, Nz)
    Fk = Ck_Fk[total_Ck_size:].reshape(6, Ny, Nx, Nz)

    # dCk_s_dt = vmap(
    #     Hermite_Fourier_system,
    #     in_axes=(None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, 0)
    # )(Ck, Fk, kx_grid, ky_grid, kz_grid, Lx, Ly, Lz, nu, D, alpha_s, u_s, qs, Omega_cs, Nn, Nm, Np, jnp.arange(Nn * Nm * Np * Ns))
    
    partial_Hermite_Fourier_system = partial(
        Hermite_Fourier_system,
        Ck, Fk, kx_grid, ky_grid, kz_grid, Lx, Ly, Lz, nu, D, alpha_s, u_s, qs, Omega_cs, Nn, Nm, Np)
    sharded_fun = jit(shard_map(vmap(partial_Hermite_Fourier_system), mesh, in_specs=spec, out_specs=spec, check_rep=False))
    indices_sharded = device_put(jnp.arange(Nn * Nm * Np * Ns), sharding)
    dCk_s_dt = sharded_fun(indices_sharded)

    nabla = jnp.array([kx_grid / Lx, ky_grid / Ly, kz_grid / Lz])
    dBk_dt = -1j * cross_product(nabla, Fk[:3])
    
    current = plasma_current(qs, alpha_s, u_s, Ck, Nn, Nm, Np, Ns)
    dEk_dt = 1j * cross_product(nabla, Fk[3:]) - current / Omega_cs[0]

    dFk_dt = jnp.concatenate([dEk_dt, dBk_dt], axis=0)
    dy_dt  = jnp.concatenate([dCk_s_dt.reshape(-1), dFk_dt.reshape(-1)])
    return dy_dt

@partial(jit, static_argnames=['Nx', 'Ny', 'Nz', 'Nn', 'Nm', 'Np', 'Ns', 'timesteps', 'solver'])
def simulation(input_parameters={}, Nx=33, Ny=1, Nz=1, Nn=20, Nm=1, Np=1, Ns=2, timesteps=200, solver=Dopri5):
    """
    Simulates the Vlasov-Maxwell system using spectral methods.
    This function initializes simulation parameters, sets up initial conditions,
    and solves the system of ordinary differential equations (ODEs) representing
    the Vlasov-Maxwell equations. The solution is returned as time-evolving
    coefficients for the distribution function (Ck) and electromagnetic fields (Fk).
    Args:
        input_parameters (dict, optional): Dictionary of user-defined simulation parameters.
        Nx (int, optional): Number of grid points in the x-direction. Default is 33.
        Ny (int, optional): Number of grid points in the y-direction. Default is 1.
        Nz (int, optional): Number of grid points in the z-direction. Default is 1.
        Nn (int, optional): Number of velocity space harmonics. Default is 20.
        Nm (int, optional): Number of azimuthal harmonics. Default is 1.
        Np (int, optional): Number of poloidal harmonics. Default is 1.
        Ns (int, optional): Number of particle species. Default is 2.
        timesteps (int, optional): Number of time steps for the simulation. Default is 200.
    Returns:
        tuple: A tuple containing:
            - Ck (jnp.ndarray): Time-evolving coefficients for the distribution function.
            - Fk (jnp.ndarray): Time-evolving coefficients for the electromagnetic fields.
            - sol.ts (jnp.ndarray): Array of time points corresponding to the solution.
    Notes:
        - The simulation uses the `diffeqsolve` function to integrate the ODEs.
        - The solution is reshaped to separate the coefficients for the distribution
          function (Ck) and the electromagnetic fields (Fk).
        - The function relies on JAX for numerical computations and efficient array operations.
    """
    
    # **Initialize simulation parameters**
    parameters = initialize_simulation_parameters(input_parameters, Nx, Ny, Nz, Nn, Nm, Np, Ns, timesteps)

    # Combine initial conditions.
    initial_conditions = jnp.concatenate([parameters["Ck_0"].flatten(), parameters["Fk_0"].flatten()])

    # Define the time array for data output.
    time = jnp.linspace(0, parameters["t_max"], timesteps)
    
    # Arguments for the ODE system.
    args = (parameters["qs"], parameters["nu"], parameters["D"], parameters["Omega_cs"], parameters["alpha_s"],
            parameters["u_s"], parameters["Lx"], parameters["Ly"], parameters["Lz"],
            parameters["kx_grid"], parameters["ky_grid"], parameters["kz_grid"])

    # Solve the ODE system
    ode_system_partial = partial(ode_system, Nx, Ny, Nz, Nn, Nm, Np, Ns)
    sol = diffeqsolve(
        ODETerm(ode_system_partial), solver=solver(),
        stepsize_controller=PIDController(rtol=parameters["ode_tolerance"], atol=parameters["ode_tolerance"]),
        t0=0, t1=parameters["t_max"], dt0=parameters["t_max"]/timesteps,
        y0=initial_conditions, args=args, saveat=SaveAt(ts=time),
        max_steps=100000, progress_meter=TqdmProgressMeter())
    
    ## Idea: take the eigenvalues of ODE_system to determine the stability of the system.
    
    # Reshape the solution to extract Ck and Fk
    Ck = sol.ys[:,:(-6 * Nx * Ny * Nz)].reshape(len(sol.ts), Ns * Nn * Nm * Np, Ny, Nx, Nz)
    Fk = sol.ys[:,(-6 * Nx * Ny * Nz):].reshape(len(sol.ts), 6, Ny, Nx, Nz)
    
    # Set n = 0, k = 0 mode to zero to get array with time evolution of perturbation.
    dCk = Ck.at[:, 0, 0, 1, 0].set(0)
    dCk = dCk.at[:, Nn * Nm * Np, 0, 1, 0].set(0)
    
    # Output results
    temporary_output = {"Ck": Ck, "Fk": Fk, "time": time, "dCk": dCk}
    output = {**temporary_output, **parameters}
    diagnostics(output)
    return output


import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.numpy.fft import ifftn, ifftshift
from jax.scipy.special import factorial
from orthax.hermite import hermval
from jax import jit, vmap
from functools import partial

__all__ = ['generate_Hermite_basis', 'generate_Hermite_term_vmap', 'inverse_HF_transform']

# Generate Hermite polynomial values for all orders from 0 to N-1
def generate_Hermite_basis(N, x):
    def compute_herm(n):
        c = jnp.zeros((N,))
        c = c.at[n].set(1.0)
        return hermval(x, c)
    return vmap(compute_herm)(jnp.arange(N))  # Shape: (N, ..., ...)

@partial(jit, static_argnames=["Nn", "Nm", "Np"])
def generate_Hermite_term_vmap(C, Herm_x, Herm_y, Herm_z, Nn, Nm, Np, xi_x, xi_y, xi_z):
    N_modes = Nn * Nm * Np
    index = jnp.arange(N_modes)

    # Precompute all combinations (n, m, p)
    p = index // (Nn * Nm)
    m = (index % (Nn * Nm)) // Nn
    n = index % Nn

    def compute_term(i, n, m, p):
        C_exp = C[:, i, :, :, :, None, None, None]  # [Nt, Ny, Nx, Nz, 1, 1, 1]
        Hx = Herm_x[n]  # [Nvy, Nvx, Nvz]
        Hy = Herm_y[m]
        Hz = Herm_z[p]
        Herm = Hx * Hy * Hz * jnp.exp(-(xi_x**2 + xi_y**2 + xi_z**2))
        norm = jnp.sqrt(jnp.pi**3) * (2 ** ((n + m + p) / 2)) * jnp.sqrt(factorial(n) * factorial(m) * factorial(p))
        return C_exp * Herm[None, None, None, None, :, :, :] / norm

    f_terms = vmap(compute_term, in_axes=(0, 0, 0, 0))(index, n, m, p)
    return jnp.sum(f_terms, axis=0)

@partial(jit, static_argnames=['Nn', 'Nm', 'Np'])
def inverse_HF_transform(Ck, Nn, Nm, Np, xi_x, xi_y, xi_z):
    C = ifftn(ifftshift(Ck, axes=(-3, -2, -1)), axes=(-3, -2, -1)).real

    # Precompute Hermite functions up to desired order
    Herm_x = generate_Hermite_basis(Nn, xi_x)  # [Nn, Nvy, Nvx, Nvz]
    Herm_y = generate_Hermite_basis(Nm, xi_y)  # [Nm, Nvy, Nvx, Nvz]
    Herm_z = generate_Hermite_basis(Np, xi_z)  # [Np, Nvy, Nvx, Nvz]

    f = generate_Hermite_term_vmap(C, Herm_x, Herm_y, Herm_z, Nn, Nm, Np, xi_x, xi_y, xi_z)
    return f


import jax.numpy as jnp
from jax import vmap, jit
from functools import partial
from jax.lax import dynamic_slice
from jax.scipy.signal import convolve

__all__ = ['plasma_current', 'collision', 'Hermite_Fourier_system', 'diffusion']

@partial(jit, static_argnames=['Nn', 'Nm', 'Np', 'Ns'])
def plasma_current(qs, alpha_s, u_s, Ck, Nn, Nm, Np, Ns):
    """
    Computes the Ampère-Maxwell current from spectral coefficients for multiple species.
    
    Parameters
    ----------
    qs : jnp.ndarray, shape (Ns,)
        Charges of the species.
    alpha_s : jnp.ndarray, shape (3 * Ns,)
        Scaling factors for each species
    u_s : jnp.ndarray, shape (3 * Ns,)
        Velocity components for each species, packed like alpha_s.
    Ck : jnp.ndarray, shape (Ns * Nn * Nm * Np, Nx, Ny, Nz)
        Spectral coefficients.
    Nn, Nm, Np : int
        Number of Hermite modes in x, y, and z respectively.
    Ns : int
        Number of species.

    Returns
    -------
    jnp.ndarray, shape (3, Nx, Ny, Nz)
        The total Ampère-Maxwell current (Jx, Jy, Jz).
    """
    # Reshape Ck into structured Hermite-Fourier coefficients: (Ns, Nn, Nm, Np, Nx, Ny, Nz)
    Ck = Ck.reshape(Ns, Nn, Nm, Np, *Ck.shape[-3:])
    
    # Reshape alpha and velocity
    alpha = alpha_s.reshape(Ns, 3)
    u = u_s.reshape(Ns, 3)

    # Grab the modes we need (0,1,1,1) for jx, jy, jz contributions
    C0 = Ck[:, 0, 0, 0]  # shape: (Ns, Nx, Ny, Nz)
    C1 = Ck[:, 1, 0, 0] if Nn > 1 else jnp.zeros_like(C0)
    Cn = Ck[:, 0, 1, 0] if Nm > 1 else jnp.zeros_like(C0)
    Cnm = Ck[:, 0, 0, 1] if Np > 1 else jnp.zeros_like(C0)

    # Pull out alpha and u components
    a0, a1, a2 = alpha[:, 0], alpha[:, 1], alpha[:, 2]
    u0, u1, u2 = u[:, 0], u[:, 1], u[:, 2]
    q = qs

    # Compute terms
    pre = q * a0 * a1 * a2  # shape: (Ns,)

    term1 = (1.0 / jnp.sqrt(2.0)) * jnp.stack([a0[:, None, None, None] * C1,
                                               a1[:, None, None, None] * Cn,
                                               a2[:, None, None, None] * Cnm], axis=0)
    term2 = jnp.stack([u0[:, None, None, None] * C0,
                       u1[:, None, None, None] * C0,
                       u2[:, None, None, None] * C0], axis=0)

    # Final current per species: shape (3, Ns, Nx, Ny, Nz)
    J_species = (term1 + term2) * pre[None, :, None, None, None]

    # Sum over species → shape: (3, Nx, Ny, Nz)
    return jnp.sum(J_species, axis=1)

@jit
def collision(Nn, Nm, Np, n, m, p):
    """
    Computes the scalar collision term for the Vlasov-Maxwell system using 
    a symmetric form over the Hermite-Fourier indices (n, m, p). 
    The contribution from each index is included only if the corresponding 
    mode count (Nn, Nm, Np) exceeds 3. The result is a differentiable sum 
    over valid contributions.
    Args:
        Nn (int): Number of grid points in the n direction.
        Nm (int): Number of grid points in the m direction.
        Np (int): Number of grid points in the p direction.
        n (int): Index in the n direction.
        m (int): Index in the m direction.
        p (int): Index in the p direction.
    Returns:
        jnp.ndarray: Scalar collision term.
    """
    N = jnp.array([Nn, Nm, Np], dtype=jnp.float32)
    idx = jnp.array([n, m, p], dtype=jnp.float32)
    # Avoid division by zero by masking invalid terms
    def safe_term(Nj, nj):
        term = nj * (nj - 1) * (nj - 2)
        denom = (Nj - 1) * (Nj - 2) * (Nj - 3)
        return jnp.where(Nj > 3, term / denom, 0.0)
    terms = vmap(safe_term)(N, idx)
    return jnp.sum(terms)

@jit
def diffusion(kx, ky, kz, D, exponent=2):
    """
    Computes the diffusion term for the Vlasov-Maxwell system.

    Parameters
    ----------
    kx, ky, kz : jnp.ndarray
        Wavevector components in the x, y, and z directions.
    D : float
        Diffusion coefficient.

    Returns
    -------
    jnp.ndarray
        Diffusion term proportional to kx**2 + ky**2 + kz**2.
    """
    k_squared = kx**exponent + ky**exponent + kz**exponent
    return -D * k_squared

@partial(jit, static_argnames=['Nn', 'Nm', 'Np'])
def Hermite_Fourier_system(Ck, Fk, kx_grid, ky_grid, kz_grid, Lx, Ly, Lz, nu, D, alpha_s, u_s, qs, Omega_cs, Nn, Nm, Np, index):
    """
    Computes the time derivative of a single Hermite-Fourier coefficient Ck[n, m, p] for species s
    in a Vlasov-Maxwell spectral solver using a Hermite-Fourier basis.

    Returns
    -------
    dCk_s_dt : jax.Array, shape (Nx, Ny, Nz)
        Time derivative of the Hermite-Fourier coefficient Ck[n, m, p] for species s.
    """
    
    # Species. s = 0 corresponds to electrons and s = 1 corresponds to ions.
    s = jnp.floor(index / (Nn * Nm * Np)).astype(jnp.int32)
    
    # Indices below represent order of Hermite polynomials (they identify the Hermite-Fourier coefficients Ck[n, p, m]).
    p = jnp.floor((index - s * Nn * Nm * Np) / (Nn * Nm)).astype(jnp.int32)
    m = jnp.floor((index - s * Nn * Nm * Np - p * Nn * Nm) / Nn).astype(jnp.int32)
    n = (index - s * Nn * Nm * Np - p * Nn * Nm - m * Nn).astype(jnp.int32)
    
    # Define u, alpha, charge, and gyrofrequency depending on species.
    u = dynamic_slice(u_s, (s * 3,), (3,))
    alpha = dynamic_slice(alpha_s, (s * 3,), (3,))
    q, Omega_c = qs[s], Omega_cs[s]
    
    # Define terms to be used in ODEs below.
    Ck_aux_x = (jnp.sqrt(m * p) * (alpha[2] / alpha[1] - alpha[1] / alpha[2]) * Ck[n + (m-1) * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) * jnp.sign(p) + 
        jnp.sqrt(m * (p + 1)) * (alpha[2] / alpha[1]) * Ck[n + (m-1) * Nn + (p+1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) * jnp.sign(Np - p - 1) - 
        jnp.sqrt((m + 1) * p) * (alpha[1] / alpha[2]) * Ck[n + (m+1) * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) * jnp.sign(Nm - m - 1) + 
        jnp.sqrt(2 * m) * (u[2] / alpha[1]) * Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) - 
        jnp.sqrt(2 * p) * (u[1] / alpha[2]) * Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p)) 

    Ck_aux_y = (jnp.sqrt(n * p) * (alpha[0] / alpha[2] - alpha[2] / alpha[0]) * Ck[n-1 + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(p) + 
        jnp.sqrt((n + 1) * p) * (alpha[0] / alpha[2]) * Ck[n+1 + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) * jnp.sign(Nn - n - 1) - 
        jnp.sqrt(n * (p + 1)) * (alpha[2] / alpha[0]) * Ck[n-1 + m * Nn + (p+1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(Np - p - 1) + 
        jnp.sqrt(2 * p) * (u[0] / alpha[2]) * Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) - 
        jnp.sqrt(2 * n) * (u[2] / alpha[0]) * Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n))
    
    Ck_aux_z = (jnp.sqrt(n * m) * (alpha[1] / alpha[0] - alpha[0] / alpha[1]) * Ck[n-1 + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(m) + 
        jnp.sqrt(n * (m + 1)) * (alpha[1] / alpha[0]) * Ck[n-1 + (m+1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(Nm - m - 1) - 
        jnp.sqrt((n + 1) * m) * (alpha[0] / alpha[1]) * Ck[n+1 + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) * jnp.sign(Nn - n - 1) + 
        jnp.sqrt(2 * n) * (u[1] / alpha[0]) * Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) - 
        jnp.sqrt(2 * m) * (u[0] / alpha[1]) * Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m))

    Col  = -nu * collision(Nn, Nm, Np, n, m, p)    * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    
    Diff = diffusion(kx_grid, ky_grid, kz_grid, D) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
        
    # ODEs for Hermite-Fourier coefficients.
    # Closure is achieved by setting to zero coefficients with index out of range.
    dCk_s_dt = (- (kx_grid * 1j / Lx) * alpha[0] * (
        jnp.sqrt((n + 1) / 2) * Ck[n+1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(Nn - n - 1) +
        jnp.sqrt(n / 2) * Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) +
        (u[0] / alpha[0]) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    ) - (ky_grid * 1j / Ly) * alpha[1] * (
        jnp.sqrt((m + 1) / 2) * Ck[n + (m+1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(Nm - m - 1) +
        jnp.sqrt(m / 2) * Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) +
        (u[1] / alpha[1]) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    ) - (kz_grid * 1j / Lz) * alpha[2] * (
        jnp.sqrt((p + 1) / 2) * Ck[n + m * Nn + (p+1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(Np - p - 1) +
        jnp.sqrt(p / 2) * Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) +
        (u[2] / alpha[2]) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    ) + q * Omega_c * (
        (jnp.sqrt(2 * n) / alpha[0]) * convolve(Fk[0, ...], Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n), mode='same') +
        (jnp.sqrt(2 * m) / alpha[1]) * convolve(Fk[1, ...], Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m), mode='same') +
        (jnp.sqrt(2 * p) / alpha[2]) * convolve(Fk[2, ...], Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p), mode='same')
    ) + q * Omega_c * (
        convolve(Fk[3, ...], Ck_aux_x, mode='same') + 
        convolve(Fk[4, ...], Ck_aux_y, mode='same') + 
        convolve(Fk[5, ...], Ck_aux_z, mode='same')
    ) + Col + Diff)
    
    return dCk_s_dt


import sys

sys.path.append(r'/Users/csvega/Desktop/Madison/Code/Simulations')
sys.path.append(r'/Users/csvega/Desktop/Madison/Code/Vlasov-Maxwell_Spectral_Solver/Vlasov-MaxwellSpectralSolver')

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.numpy.fft import fftn, fftshift
from jax.scipy.special import factorial
from jax.scipy.integrate import trapezoid
from orthax.hermite import hermval3d

def Hermite(n, x):
    """
    I have to add docstrings!
    """
    
    n = n.astype(int) # Ensure that n is an integer.
    
    # Add next term in Hermite polynomial. Body function of fori_loop below.
    def add_Hermite_term(m, partial_sum):
        return partial_sum + ((-1)**m / (factorial(m) * factorial(n - 2*m))) * (2*x)**(n - 2*m)
    
    # Return Hermite polynomial of order n.
    return factorial(n) * jax.lax.fori_loop(0, (n // 2) + 1, add_Hermite_term, jnp.zeros_like(x))


def generate_Hermite_function(xi_x, xi_y, xi_z, Nn, Nm, indices):
    """
    I have to add docstrings!
    """
    
    # Indices below represent order of Hermite polynomials.
    p = jnp.floor(indices / (Nn * Nm)).astype(int)
    m = jnp.floor((indices - p * Nn * Nm) / Nn).astype(int)
    n = (indices - p * Nn * Nm - m * Nn).astype(int)
    
    # Generate element of AW Hermite basis in 3D space.
    Hermite_basis = (Hermite(n, xi_x) * Hermite(m, xi_y) * Hermite(p, xi_z) * 
                jnp.exp(-(xi_x**2 + xi_y**2 + xi_z**2)) / 
                jnp.sqrt((jnp.pi)**3 * 2**(n + m + p) * factorial(n) * factorial(m) * factorial(p)))
    
    return Hermite_basis

def compute_C_nmp(f, alpha, u, Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, indices):
    """
    I have to add docstrings!
    """
    
    # Indices below represent order of Hermite polynomials.
    p = jnp.floor(indices / (Nn * Nm)).astype(int)
    m = jnp.floor((indices - p * Nn * Nm) / Nn).astype(int)
    n = (indices - p * Nn * Nm - m * Nn).astype(int)

    # Generate 6D space for particle distribution function f.
    x = jnp.linspace(0, Lx, Nx)
    y = jnp.linspace(0, Ly, Ny)
    z = jnp.linspace(0, Lz, Nz)
    vx = jnp.linspace(-5 * alpha[0] + u[0], 5 * alpha[0] + u[0], 40)
    vy = jnp.linspace(-5 * alpha[1] + u[1], 5 * alpha[1] + u[1], 40)
    vz = jnp.linspace(-5 * alpha[2] + u[2], 5 * alpha[2] + u[2], 40)
      
    def add_C_nmp(i, C_nmp):
        ivx = jnp.floor(i / (5 ** 2)).astype(int)
        ivy = jnp.floor((i - ivx * 5 ** 2) / 5).astype(int)
        ivz = (i - ivx * 5 ** 2 - ivy * 5).astype(int)
        
        vx_slice = jax.lax.dynamic_slice(vx, (ivx * 8,), (8,))
        vy_slice = jax.lax.dynamic_slice(vy, (ivy * 8,), (8,))
        vz_slice = jax.lax.dynamic_slice(vz, (ivz * 8,), (8,))
        
        X, Y, Z, Vx, Vy, Vz = jnp.meshgrid(x, y, z, vx_slice, vy_slice, vz_slice, indexing='xy')

        # Define variables for Hermite polynomials.
        xi_x = (Vx - u[0]) / alpha[0]
        xi_y = (Vy - u[1]) / alpha[1]
        xi_z = (Vz - u[2]) / alpha[2]

        # Compute coefficients of Hermite decomposition of 3D velocity space.
        return C_nmp + (trapezoid(trapezoid(trapezoid(
                (f(X, Y, Z, Vx, Vy, Vz) * Hermite(n, xi_x) * Hermite(m, xi_y) * Hermite(p, xi_z)) /
                jnp.sqrt(factorial(n) * factorial(m) * factorial(p) * 2 ** (n + m + p)),
                (vy_slice - u[0]) / alpha[0], axis=-3), (vx_slice - u[1]) / alpha[1], axis=-2), (vz_slice - u[2]) / alpha[2], axis=-1))
                
    Nv = 125
        
    return jax.lax.fori_loop(0, Nv, add_C_nmp, jnp.zeros((Ny, Nx, Nz)))


def initialize_system_xv(B, E, f1, f2, Omega_ce, mi_me, alpha_s, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns):
    """
    I have to add docstrings!
    """
        
    # Hermite decomposition of dsitribution funcitons.
    C1_0 = (jax.vmap(
        compute_C_nmp, in_axes=(
            None, None, None, None, None, None, None, None, None, None, None, None, 0))
        (f1, alpha_s[:3], u_s[:3], Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, jnp.arange(Nn * Nm * Np)))
    C2_0 = (jax.vmap(
        compute_C_nmp, in_axes=(
            None, None, None, None, None, None, None, None, None, None, None, None, 0))
        (f2, alpha_s[3:], u_s[3:], Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, jnp.arange(Nn * Nm * Np)))

    # Combine Ce_0 and Ci_0 into single array and compute the fast Fourier transform.
    
    C1k_0 = fftshift(fftn(C1_0, axes=(-3, -2, -1)), axes=(-3, -2, -1))
    C2k_0 = fftshift(fftn(C2_0, axes=(-3, -2, -1)), axes=(-3, -2, -1))

    
    Ck_0 = jnp.concatenate([C1k_0, C2k_0])
    # Ck_0 = fftshift(fftn(C_0, axes=(-3, -2, -1)), axes=(-3, -2, -1))
    
    ############################################################################################################   
    # Attempt to generalize to more than two species (work in progress).
    
    # B, E, f = density_perturbation_1D(Lx, Omega_ce, mi_me)
    
    # C_0 = jnp.zeros((Ns * Nn * Nm * Np))
    # for s in jnp.arange(Ns):
    #     C_s = (jax.vmap(compute_C_nmp, in_axes=(
    #         None, None, None, None, None, None, None, None, None, None, None, None, 0))
    #     (f[s], alpha_s[(s * 3):((s + 1) * 3)], u_s[(s * 3):((s + 1) * 3)], Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, jnp.arange(Nn * Nm * Np)))
        
    #     C_0.at[(s * Ns * Nn * Nm * Np):((s + 1) * Ns * Nn * Nm * Np)].set(C_s)
    
    ############################################################################################################
    
    # Define 3D grid for functions E(x, y, z) and B(x, y, z).
    x = jnp.linspace(0, Lx, Nx)
    y = jnp.linspace(0, Ly, Ny)
    z = jnp.linspace(0, Lz, Nz)
    X, Y, Z = jnp.meshgrid(x, y, z, indexing='xy')
    
    # Combine E and B into single array and compute the fast Fourier transform.
    
    Ek_0 = fftshift(fftn(E(X, Y, Z), axes=(-3, -2, -1)), axes=(-3, -2, -1))
    Bk_0 = fftshift(fftn(B(X, Y, Z), axes=(-3, -2, -1)), axes=(-3, -2, -1))

    Fk_0 = jnp.concatenate([Ek_0, Bk_0])
    # Fk_0 = fftshift(fftn(F_0, axes=(-3, -2, -1)), axes=(-3, -2, -1))
    
    return Ck_0, Fk_0


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 23 10:55:45 2024

@author: cristian
"""

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.scipy.signal import convolve
from jax.scipy.special import factorial
from jax.numpy.fft import fftn, ifftn, fftshift, ifftshift
from jax.experimental.ode import odeint
from diffrax import diffeqsolve, Dopri5, Tsit5, Kvaerno3, ImplicitEuler, ODETerm, SaveAt, ConstantStepSize, PIDController
from functools import partial


def cross_product(k_vec, F_vec):
    """
    I have to add docstrings!
    """
    
    # Separate vectors into x, y, z components.
    kx, ky, kz = k_vec
    Fx, Fy, Fz = F_vec
    
    # Compute the cross product k x F.
    result_x = ky * Fz - kz * Fy
    result_y = kz * Fx - kx * Fz
    result_z = kx * Fy - ky * Fx

    return jnp.array([result_x, result_y, result_z])


def collision(Nn, Nm, Np, n, m, p):
    # Pack inputs into arrays
    Nnmp = jnp.array([Nn, Nm, Np])
    nmp = jnp.array([n, m, p])
    
    # Sort by descending N values
    indices = jnp.argsort(-Nnmp)
    sorted_Nnmp = Nnmp[indices]
    sorted_nmp = nmp[indices]
    
    # Helper function to calculate collision term for a single component
    def collision_term(Nj, nj):
        return nj * (nj - 1) * (nj - 2) / ((Nj - 1) * (Nj - 2) * (Nj - 3))
    
    # Calculate all possible terms in a vectorized way
    valid_indices = sorted_Nnmp > 3
    
    # Calculate the sum for valid indices
    result = jnp.sum(jnp.where(valid_indices, collision_term(sorted_Nnmp, sorted_nmp), 0.0))
    
    # Return 0 if no valid indices
    return jax.lax.cond(
        jnp.any(valid_indices),
        lambda _: result,
        lambda _: jnp.array(0.0),
        operand=None
    )

def compute_dCk_s_dt(Ck, Fk, kx_grid, ky_grid, kz_grid, Lx, Ly, Lz, nu, alpha_s, u_s, qs, Omega_cs, Nn, Nm, Np, index):
    """
    I have to add docstrings!
    """
    
    # Species. s = 0 corresponds to electrons and s = 1 corresponds to ions.
    s = jnp.floor(index / (Nn * Nm * Np)).astype(int)
    
    # Indices below represent order of Hermite polynomials (they identify the Hermite-Fourier coefficients Ck[n, p, m]).
    p = jnp.floor((index - s * Nn * Nm * Np) / (Nn * Nm)).astype(int)
    m = jnp.floor((index - s * Nn * Nm * Np - p * Nn * Nm) / Nn).astype(int)
    n = (index - s * Nn * Nm * Np - p * Nn * Nm - m * Nn).astype(int)
    
    # Define u, alpha, charge, and gyrofrequency depending on species.
    u = jax.lax.dynamic_slice(u_s, (s * 3,), (3,))
    alpha = jax.lax.dynamic_slice(alpha_s, (s * 3,), (3,))
    q, Omega_c = qs[s], Omega_cs[s]
    
    # Define terms to be used in ODEs below.
    Ck_aux_x = (jnp.sqrt(m * p) * (alpha[2] / alpha[1] - alpha[1] / alpha[2]) * Ck[n + (m-1) * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) * jnp.sign(p) + 
        jnp.sqrt(m * (p + 1)) * (alpha[2] / alpha[1]) * Ck[n + (m-1) * Nn + (p+1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) * jnp.sign(Np - p - 1) - 
        jnp.sqrt((m + 1) * p) * (alpha[1] / alpha[2]) * Ck[n + (m+1) * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) * jnp.sign(Nm - m - 1) + 
        jnp.sqrt(2 * m) * (u[2] / alpha[1]) * Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) - 
        jnp.sqrt(2 * p) * (u[1] / alpha[2]) * Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p)) 

    Ck_aux_y = (jnp.sqrt(n * p) * (alpha[0] / alpha[2] - alpha[2] / alpha[0]) * Ck[n-1 + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(p) + 
        jnp.sqrt((n + 1) * p) * (alpha[0] / alpha[2]) * Ck[n+1 + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) * jnp.sign(Nn - n - 1) - 
        jnp.sqrt(n * (p + 1)) * (alpha[2] / alpha[0]) * Ck[n-1 + m * Nn + (p+1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(Np - p - 1) + 
        jnp.sqrt(2 * p) * (u[0] / alpha[2]) * Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) - 
        jnp.sqrt(2 * n) * (u[2] / alpha[0]) * Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n))
    
    Ck_aux_z = (jnp.sqrt(n * m) * (alpha[1] / alpha[0] - alpha[0] / alpha[1]) * Ck[n-1 + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(m) + 
        jnp.sqrt(n * (m + 1)) * (alpha[1] / alpha[0]) * Ck[n-1 + (m+1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) * jnp.sign(Nm - m - 1) - 
        jnp.sqrt((n + 1) * m) * (alpha[0] / alpha[1]) * Ck[n+1 + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) * jnp.sign(Nn - n - 1) + 
        jnp.sqrt(2 * n) * (u[1] / alpha[0]) * Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) - 
        jnp.sqrt(2 * m) * (u[0] / alpha[1]) * Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m))
    
    # Define "unphysical" collision operator to eliminate recurrence.

    Col = -nu * collision(Nn,Nm,Np,n,m,p) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
        
    # Define ODEs for Hermite-Fourier coefficients.
    # Clossure is achieved by setting to zero coefficients with index out of range.
    dCk_s_dt = (- (kx_grid * 1j / Lx) * alpha[0] * (
        jnp.sqrt((n + 1) / 2) * Ck[n+1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(Nn - n - 1) +
        jnp.sqrt(n / 2) * Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n) +
        (u[0] / alpha[0]) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    ) - (ky_grid * 1j / Ly) * alpha[1] * (
        jnp.sqrt((m + 1) / 2) * Ck[n + (m+1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(Nm - m - 1) +
        jnp.sqrt(m / 2) * Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m) +
        (u[1] / alpha[1]) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    ) - (kz_grid * 1j / Lz) * alpha[2] * (
        jnp.sqrt((p + 1) / 2) * Ck[n + m * Nn + (p+1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(Np - p - 1) +
        jnp.sqrt(p / 2) * Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p) +
        (u[2] / alpha[2]) * Ck[n + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...]
    ) + q * Omega_c * (
        (jnp.sqrt(2 * n) / alpha[0]) * convolve(Fk[0, ...], Ck[n-1 + m * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(n), mode='same') +
        (jnp.sqrt(2 * m) / alpha[1]) * convolve(Fk[1, ...], Ck[n + (m-1) * Nn + p * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(m), mode='same') +
        (jnp.sqrt(2 * p) / alpha[2]) * convolve(Fk[2, ...], Ck[n + m * Nn + (p-1) * Nn * Nm + s * Nn * Nm * Np, ...] * jnp.sign(p), mode='same')
    ) + q * Omega_c * (
        convolve(Fk[3, ...], Ck_aux_x, mode='same') + 
        convolve(Fk[4, ...], Ck_aux_y, mode='same') + 
        convolve(Fk[5, ...], Ck_aux_z, mode='same')
    ) + Col)
    
    return dCk_s_dt


def ampere_maxwell_current(qs, alpha_s, u_s, Ck, Nn, Nm, Np, Ns):
    """
    I have to add docstrings!
    """
      
    # Add next term in the current. Body function of fori_loop below.
    def add_current_term(s, partial_sum):
        return partial_sum + qs[s] * alpha_s[s * 3] * alpha_s[s * 3 + 1] * alpha_s[s * 3 + 2] * (
            (1 / jnp.sqrt(2)) * jnp.array([alpha_s[s * 3] * Ck[s * Nn * Nm * Np + 1, ...] * jnp.sign(Nn - 1),
                                           alpha_s[s * 3 + 1] * Ck[s * Nn * Nm * Np + Nn, ...] * jnp.sign(Nm - 1),
                                           alpha_s[s * 3 + 2] * Ck[s * Nn * Nm * Np + Nn * Nm, ...] * jnp.sign(Np - 1)]) + 
                                jnp.array([u_s[s * 3] * Ck[s * Nn * Nm * Np, ...],
                                           u_s[s * 3 + 1] * Ck[s * Nn * Nm * Np, ...],
                                           u_s[s * 3 + 2] * Ck[s * Nn * Nm * Np, ...]]))
    
    # Return full current for Ns particle species.
    return jax.lax.fori_loop(0, Ns, add_current_term, jnp.zeros_like(Ck[:3, ...]))


# def ode_system(t, Ck_Fk, qs, nu, Omega_cs, alpha_s, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns):
def ode_system(t, Ck_Fk, args):

    qs, nu, Omega_cs, alpha_s, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns = args
    
    # Define wave vectors.
    kx = (jnp.arange(-Nx//2, Nx//2) + 1) * 2 * jnp.pi
    ky = (jnp.arange(-Ny//2, Ny//2) + 1) * 2 * jnp.pi
    kz = (jnp.arange(-Nz//2, Nz//2) + 1) * 2 * jnp.pi  

    # Create 3D grids of kx, ky, kz.
    kx_grid, ky_grid, kz_grid = jnp.meshgrid(kx, ky, kz, indexing='xy')
    
    # Separate between initial conditions for distribution functions (coefficients Ck)
    # and electric and magnetic fields (coefficients Fk).
    Ck = Ck_Fk[:(-6 * Nx * Ny * Nz)].reshape(Ns * Nn * Nm * Np, Ny, Nx, Nz)
    Fk = Ck_Fk[(-6 * Nx * Ny * Nz):].reshape(6, Ny, Nx, Nz)
    
    # Vectorize over n, m, p, and s to generate ODEs for all coefficients Ck.
    dCk_s_dt = (jax.vmap(
        compute_dCk_s_dt, 
        in_axes=(None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, 0))
        (Ck, Fk, kx_grid, ky_grid, kz_grid, Lx, Ly, Lz, nu, alpha_s, u_s, qs, Omega_cs, Nn, Nm, Np, jnp.arange(Nn * Nm * Np * Ns)))
    
    
    current = ampere_maxwell_current(qs, alpha_s, u_s, Ck, Nn, Nm, Np, Ns)
        
    # Generate ODEs for Bk and Ek.
    dBk_dt = - 1j * cross_product(jnp.array([kx_grid/Lx, ky_grid/Ly, kz_grid/Lz]), Fk[:3, ...])
    dEk_dt = 1j * cross_product(jnp.array([kx_grid/Lx, ky_grid/Ly, kz_grid/Lz]), Fk[3:, ...]) - (1 / Omega_cs[0]) * current
            
            # (qs[0] * alpha_s[0] * alpha_s[1] * alpha_s[2] * (
            # (1 / jnp.sqrt(2)) * jnp.array([alpha_s[0] * Ck[1, ...] * jnp.sign(Nn - 1),
            #                                alpha_s[1] * Ck[Nn + 1, ...] * jnp.sign(Nm - 1),
            #                                alpha_s[2] * Ck[Nn * Nm + 1, ...] * jnp.sign(Np - 1)]) + 
            #                     jnp.array([u_s[0] * Ck[0, ...],
            #                                u_s[1] * Ck[0, ...],
            #                                u_s[2] * Ck[0, ...]])) + \
            #                     qs[1] * alpha_s[3] * alpha_s[4] * alpha_s[5] * (
            # (1 / jnp.sqrt(2)) * jnp.array([alpha_s[3] * Ck[Nn * Nm * Np + 1, ...] * jnp.sign(Nn - 1),
            #                                alpha_s[4] * Ck[Nn + Nn * Nm * Np + 1, ...] * jnp.sign(Nm - 1),
            #                                alpha_s[5] * Ck[Nn * Nm + Nn * Nm * Np + 1, ...] * jnp.sign(Np - 1)]) + 
            #                     jnp.array([u_s[3] * Ck[Nn * Nm * Np, ...],
            #                                u_s[4] * Ck[Nn * Nm * Np, ...],
            #                                u_s[5] * Ck[Nn * Nm * Np, ...]])))

    # Combine dC/dt and dF/dt into a single array and flatten it into a 1D array for an ODE solver.
    dFk_dt = jnp.concatenate([dEk_dt, dBk_dt])
    dy_dt = jnp.concatenate([dCk_s_dt.flatten(), dFk_dt.flatten()])

    
    return dy_dt


@partial(jax.jit, static_argnums=[11, 12, 13, 14, 15, 16, 17, 19])
def VM_simulation(Ck_0, Fk_0, qs, nu, Omega_cs, alpha_s, mi_me, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns, t_max, t_steps, dt):
    

    # Combine initial conditions.
    initial_conditions = jnp.concatenate([Ck_0.flatten(), Fk_0.flatten()])

    # Define the time array for data output.
    t = jnp.linspace(0, t_max, t_steps)

    args = (qs, nu, Omega_cs, alpha_s, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns)
    
###################################################################################################    
    
    # Solve ODE system using diffrax.
    saveat = SaveAt(ts=t)
    term = ODETerm(ode_system)
    solver = Dopri5()
    # stepsize_controller = ConstantStepSize()
    # stepsize_controller = PIDController(rtol=1e-6, atol=1e-6)
    result = diffeqsolve(term, solver, t0=0, t1=t_max, dt0=dt, y0=initial_conditions, args=args, saveat=saveat, max_steps=100000)
    
    Ck = result.ys[:,:(-6 * Nx * Ny * Nz)].reshape(len(result.ts), Ns * Nn * Nm * Np, Ny, Nx, Nz)
    Fk = result.ys[:,(-6 * Nx * Ny * Nz):].reshape(len(result.ts), 6, Ny, Nx, Nz)
    
    # jnp.save('Ck.npy', np.array(Ck))
    # jnp.save('Fk.npy', np.array(Fk))
    # jnp.save('t.npy', np.array(t))

    return Ck, Fk, result.ts
    
###################################################################################################
    # # Solve the ODE system using odeint.
    
    # dy_dt = partial(ode_system, args=args)
    # result = odeint(dy_dt, initial_conditions, t)
    
    # Ck = result[:,:(-6 * Nx * Ny * Nz)].reshape(len(t), Ns * Nn * Nm * Np, Ny, Nx, Nz)
    # Fk = result[:,(-6 * Nx * Ny * Nz):].reshape(len(t), 6, Ny, Nx, Nz)
    
    # return Ck, Fk, t 



import os
number_of_processors_to_use = 5 # Parallelization, this should divide total resolution
os.environ["XLA_FLAGS"] = f'--xla_force_host_platform_device_count={number_of_processors_to_use}'
from time import time
from jax import block_until_ready
from spectrax import simulation, load_parameters, plot

# Read from input.toml
input_parameters, solver_parameters = load_parameters('input_1D_two_stream.toml')

# Simulate
start_time = time()
output = block_until_ready(simulation(input_parameters, **solver_parameters))
print(f"Runtime: {time() - start_time} seconds")

# Plot results
plot(output)

# print('Saving results...')
# jnp.savez('output_two-stream.npz', **output)

# print("Loading results...")
# output = jnp.load('output_orszag.npz')


import jax.numpy as jnp
try: import tomllib
except ModuleNotFoundError: import pip._vendor.tomli as tomllib
import diffrax
import inspect

__all__ = ["load_parameters", "initialize_simulation_parameters"]

def initialize_simulation_parameters(user_parameters={}, Nx=33, Ny=1, Nz=1, Nn=50, Nm=1, Np=1, Ns=2, timesteps=500):
    """
    Initialize the simulation parameters for a Vlasov solver with Hermite polynomials, 
    combining user-provided values with predefined defaults. This function 
    ensures all required parameters are set and automatically calculates 
    derived parameters based on the inputs.

    The function uses lambda functions to define derived parameters that 
    depend on other parameters. These lambda functions are evaluated after 
    merging user-provided parameters with the defaults, ensuring derived 
    parameters are consistent with any overrides.

    Parameters:
    ----------
    user_parameters : dict
        Dictionary containing user-specified parameters. Any parameter not provided
        will default to predefined values.

    Returns:
    -------
    parameters : dict
        Dictionary containing all simulation parameters, with user-provided values
        overriding defaults.
    """
    # Define all default parameters in a single dictionary
    default_parameters = {
        "Lx": 4 * jnp.pi,
        "Ly": 1.0,
        "Lz": 1.0,
        "mi_me": 1.0,
        "Ti_Te": 1.0,
        "qs": jnp.array([-1, -1]),
        "alpha_e": jnp.array([0.707107, 0.707107, 0.707107]),
        "alpha_s": lambda p: jnp.concatenate([
            p["alpha_e"],
            p["alpha_e"] * jnp.sqrt(p["Ti_Te"] / p["mi_me"])
        ]),
        "u_s": jnp.array([1.0, 0.0, 0.0, -1.0, 0.0, 0.0]),
        "Omega_cs": lambda p: jnp.array([1.0, 1.0 / p["mi_me"]]),
        "nu": 2.0,
        "D": 0.0,
        "t_max": 0.3,
        "kx": lambda p: 2 * jnp.pi / p["Lx"],
        "dn1": 0.001,
        "dn2": 0.001,
        "ode_tolerance": 1e-6,
        "vte": lambda p: p["alpha_s"][0] / jnp.sqrt(2),
        "vti": lambda p: p["vte"] * jnp.sqrt(1 / p["mi_me"]),
    }
    
    # Initialize distribution function as a two-stream instability
    indices = jnp.array([int((Nx-1)/2-1), int((Nx-1)/2+1)])
    dn1     = default_parameters["dn1"]
    dn2     = default_parameters["dn2"]
    alpha_e = default_parameters["alpha_e"]
    values  = (dn1 + dn2) / (2 * default_parameters["kx"](default_parameters) * default_parameters["Omega_cs"](default_parameters)[0])
    Fk_0    = jnp.zeros((6, 1, Nx, 1), dtype=jnp.complex128).at[0, 0, indices, 0].set(values)
    C10     = jnp.array([
            0 + 1j * (1 / (2 * alpha_e[0] ** 3)) * dn1,
            1 / (alpha_e[0] ** 3) + 0 * 1j,
            0 - 1j * (1 / (2 * alpha_e[0] ** 3)) * dn1
    ])
    C20     = jnp.array([
            0 + 1j * (1 / (2 * alpha_e[0] ** 3)) * dn2,
            1 / (alpha_e[0] ** 3) + 0 * 1j,
            0 - 1j * (1 / (2 * alpha_e[0] ** 3)) * dn2
    ])
    indices = jnp.array([int((Nx-1)/2-1), int((Nx-1)/2), int((Nx-1)/2+1)])
    Ck_0    = jnp.zeros((2 * Nn, 1, Nx, 1), dtype=jnp.complex128)
    Ck_0    = Ck_0.at[0,  0, indices, 0].set(C10)
    Ck_0    = Ck_0.at[Nn, 0, indices, 0].set(C20)
    
    default_parameters.update({
        "Ck_0": Ck_0, "Fk_0": Fk_0,
        "timesteps": timesteps, "Ns": Ns,
        "Nx": Nx, "Ny": Ny, "Nz": Nz,
        "Nn": Nn, "Nm": Nm, "Np": Np,
    })

    # Merge user-provided parameters into the default dictionary
    parameters = {**default_parameters, **user_parameters}
    
    # Compute derived parameters based on user-provided or default values
    for key, value in parameters.items():
        if callable(value):  # If the value is a lambda function, compute it
            parameters[key] = value(parameters)
        if isinstance(value, list):
            parameters[key] = jnp.array(value)
    kx_simulation = (jnp.arange(-Nx//2, Nx//2) + 1) * 2 * jnp.pi
    ky_simulation = (jnp.arange(-Ny//2, Ny//2) + 1) * 2 * jnp.pi
    kz_simulation = (jnp.arange(-Nz//2, Nz//2) + 1) * 2 * jnp.pi  
    kx_grid, ky_grid, kz_grid = jnp.meshgrid(kx_simulation, ky_simulation, kz_simulation, indexing='xy')
    parameters.update({
        "kx_grid": kx_grid, "ky_grid": ky_grid, "kz_grid": kz_grid
    })

    return parameters

def load_parameters(input_file):
    """
    Load parameters from a TOML file.

    Parameters:
    ----------
    input_file : str
        Path to the TOML file containing simulation parameters.

    Returns:
    -------
    parameters : dict
        Dictionary containing simulation parameters.
    """
    parameters = tomllib.load(open(input_file, "rb"))
    input_parameters = parameters['input_parameters']
    solver_parameters = parameters['solver_parameters']

    def get_solver_class(name: str):
        for cls_name, cls in inspect.getmembers(diffrax, inspect.isclass):
            if issubclass(cls, diffrax.AbstractSolver) and cls is not diffrax.AbstractSolver and cls_name == name: return cls
        raise ValueError(f"Solver '{name}' is not supported. Choose from Diffrax solvers.")
    solver_parameters["solver"] = get_solver_class(solver_parameters.get("solver", "Tsit5"))
    
    return input_parameters, solver_parameters



import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax import vmap, jit
from jax.numpy.fft import fftn, fftshift
from jax.scipy.special import factorial
from jax.scipy.integrate import trapezoid
from functools import partial

__all__ = ['Hermite', 'generate_Hermite_function', 'compute_C_nmp', 'initialize_xv']

# Parallelize the simulation using JAX
from jax.sharding import PartitionSpec as P
from jax.experimental.shard_map import shard_map
from jax import devices, make_mesh, NamedSharding, device_put
mesh = make_mesh((len(devices()),), ("batch"))
spec = P("batch")
sharding = NamedSharding(mesh, spec)

@jit
def Hermite(n, x):
    def body(m, acc):
        coeff = (-1.0)**m / (factorial(m) * factorial(n - 2 * m))
        term = coeff * (2 * x)**(n - 2 * m)
        return acc + term
    return factorial(n) * jax.lax.fori_loop(0, (n // 2) + 1, body, jnp.zeros_like(x))

@jit
def generate_Hermite_function(xi_x, xi_y, xi_z, Nn, Nm, indices):
    p = indices // (Nn * Nm)
    m = (indices % (Nn * Nm)) // Nn
    n = indices % Nn
    norm = jnp.sqrt(jnp.pi**3 * 2**(n + m + p) * factorial(n) * factorial(m) * factorial(p))
    return (
        Hermite(n, xi_x) * Hermite(m, xi_y) * Hermite(p, xi_z) *
        jnp.exp(-(xi_x**2 + xi_y**2 + xi_z**2)) / norm
    )

@partial(jit, static_argnames=['f', 'Nn', 'Nm', 'Np', 'Nx', 'Ny', 'Nz', 'Nv', 'nvxyz', 'max_min_v_factor'])
def compute_C_nmp(f, alpha, u, Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, Nv, nvxyz, max_min_v_factor, indices):
    # Indices below represent order of Hermite polynomials.
    p = jnp.floor(indices / (Nn * Nm)).astype(int)
    m = jnp.floor((indices - p * Nn * Nm) / Nn).astype(int)
    n = (indices - p * Nn * Nm - m * Nn).astype(int)
    # Generate 6D space for particle distribution function f.
    x = jnp.linspace(0, Lx, Nx)
    y = jnp.linspace(0, Ly, Ny)
    z = jnp.linspace(0, Lz, Nz)
    vx = jnp.linspace(-max_min_v_factor * alpha[0] + u[0], max_min_v_factor * alpha[0] + u[0], nvxyz)
    vy = jnp.linspace(-max_min_v_factor * alpha[1] + u[1], max_min_v_factor * alpha[1] + u[1], nvxyz)
    vz = jnp.linspace(-max_min_v_factor * alpha[2] + u[2], max_min_v_factor * alpha[2] + u[2], nvxyz)
    # Precompute all 8x8x8 velocity subgrids
    def slice_1d(v):
        return jnp.stack([v[i * 8:(i + 1) * 8] for i in range(5)])
    vx_chunks = slice_1d(vx)  # shape (5, 8)
    vy_chunks = slice_1d(vy)
    vz_chunks = slice_1d(vz)
    
    xv, yv, zv = jnp.meshgrid(x, y, z, indexing='xy')  # shape (Ny, Nx, Nz)
    X = xv[..., None, None, None]
    Y = yv[..., None, None, None]
    Z = zv[..., None, None, None]
  
    @jit
    def single_C_nmp(i):
        ivx = jnp.floor(i / (5 ** 2)).astype(int)
        ivy = jnp.floor((i - ivx * 5 ** 2) / 5).astype(int)
        ivz = (i - ivx * 5 ** 2 - ivy * 5).astype(int)
        
        vx_slice = vx_chunks[ivx]
        vy_slice = vy_chunks[ivy]
        vz_slice = vz_chunks[ivz]
        
        # X, Y, Z, Vx, Vy, Vz = jnp.meshgrid(x, y, z, vx_slice, vy_slice, vz_slice, indexing='xy')
        vxv, vyv, vzv = jnp.meshgrid(vx_slice, vy_slice, vz_slice, indexing='ij')  # shape (8, 8, 8)
        Vx = vxv[None, None, None, ...]
        Vy = vyv[None, None, None, ...]
        Vz = vzv[None, None, None, ...]

        # Compute coefficients of Hermite decomposition of 3D velocity space.
        xi_x = (Vx - u[0]) / alpha[0]
        xi_y = (Vy - u[1]) / alpha[1]
        xi_z = (Vz - u[2]) / alpha[2]
        Hx = Hermite(n, xi_x)
        Hy = Hermite(m, xi_y)
        Hz = Hermite(p, xi_z)
        hermite_norm = jnp.sqrt(factorial(n) * factorial(m) * factorial(p) * 2 ** (n + m + p))
        
        # Compute the integrand for the distribution function.
        integrand = f(X, Y, Z, Vx, Vy, Vz) * Hx * Hy * Hz / hermite_norm
        
        # Compute the contribution to C_nmp from the velocity space.
        contribution = trapezoid(trapezoid(trapezoid(integrand,
                    (vy_slice - u[0]) / alpha[0], axis=-3),
                    (vx_slice - u[1]) / alpha[1], axis=-2),
                    (vz_slice - u[2]) / alpha[2], axis=-1)
        return contribution
    C_nmp_all = vmap(single_C_nmp)(jnp.arange(Nv))  # shape (Nv, Ny, Nx, Nz)
    return jnp.sum(C_nmp_all, axis=0)

@partial(jit, static_argnames=['B', 'E', 'f1', 'f2', 'Nv', 'nvxyz', 'max_min_v_factor', 'Nx', 'Ny', 'Nz', 'Nn', 'Nm', 'Np', 'Ns', 'timesteps', 'solver'])
def initialize_xv(B, E, f1, f2, Nv=55, nvxyz=40, max_min_v_factor=5, input_parameters={},
                  Nx=33, Ny=33, Nz=1, Nn=4, Nm=4, Np=4, Ns=None, timesteps=None, solver=None):
    alpha_s = input_parameters["alpha_s"]
    u_s = input_parameters["u_s"]
    Lx = input_parameters["Lx"]
    Ly = input_parameters["Ly"]
    Lz = input_parameters["Lz"]
    
    indices_sharded = jnp.arange(Nn * Nm * Np)
    indices_sharded = device_put(indices_sharded, sharding)
    
    partial_C1_0 = partial(compute_C_nmp, f1, alpha_s[:3], u_s[:3], Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, Nv, nvxyz, max_min_v_factor)
    sharded_C10_fun = jit(shard_map(vmap(partial_C1_0), mesh, in_specs=spec, out_specs=spec, check_rep=False))
    C1_0 = sharded_C10_fun(indices_sharded)
    
    partial_C2_0 = partial(compute_C_nmp, f2, alpha_s[3:], u_s[3:], Nx, Ny, Nz, Lx, Ly, Lz, Nn, Nm, Np, Nv, nvxyz, max_min_v_factor)
    sharded_C20_fun = jit(shard_map(vmap(partial_C2_0), mesh, in_specs=spec, out_specs=spec, check_rep=False))
    C2_0 = sharded_C20_fun(indices_sharded)

    # Combine Ce_0 and Ci_0 into single array and compute the fast Fourier transform.
    C1k_0 = fftshift(fftn(C1_0, axes=(-3, -2, -1)), axes=(-3, -2, -1))
    C2k_0 = fftshift(fftn(C2_0, axes=(-3, -2, -1)), axes=(-3, -2, -1))
    Ck_0 = jnp.concatenate([C1k_0, C2k_0])
    
    # Define 3D grid for functions E(x, y, z) and B(x, y, z).
    x = jnp.linspace(0, Lx, Nx)
    y = jnp.linspace(0, Ly, Ny)
    z = jnp.linspace(0, Lz, Nz)
    X, Y, Z = jnp.meshgrid(x, y, z, indexing='xy')
    
    # Combine E and B into single array and compute the fast Fourier transform.
    Ek_0 = fftshift(fftn(E(X, Y, Z), axes=(-3, -2, -1)), axes=(-3, -2, -1))
    Bk_0 = fftshift(fftn(B(X, Y, Z), axes=(-3, -2, -1)), axes=(-3, -2, -1))
    Fk_0 = jnp.concatenate([Ek_0, Bk_0])
    
    return Ck_0, Fk_0


import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.numpy.fft import ifftn, ifftshift
from jax.scipy.special import factorial
from orthax.hermite import hermval3d, hermval
from jax import debug

def Hermite(n, x):
    """
    I have to add docstrings!
    """
    
    c = jnp.zeros(n+1)
    c = c.at[-1].set(1)
    
    return hermval(x, c)
    
    
def generate_C_term(X, Y, Z, kx, ky, kz, Ck, Nx, Ny, index):
   
    # Indices below represent order of Hermite polynomials.
    nz = jnp.floor(index / (Nx * Ny)).astype(int)
    ny = jnp.floor((index - nz * Nx * Ny) / Nx).astype(int)
    nx = (index - nz * Nx * Ny - ny * Nx).astype(int)
    
    expkx = jnp.exp(1j * (kx[nx] * X + ky[ny] * Y + kz[nz] * Z))
    expkx_extended = expkx[jnp.newaxis, jnp.newaxis, :, :, :]
    
    return Ck[:, :, ny, nx, nz, jnp.newaxis, jnp.newaxis, jnp.newaxis] * expkx_extended


def generate_Hermite_term(xi_x, xi_y, xi_z, C, Nn, Nm, index):
    """
    I have to add docstrings!
    """
    
    # Indices below represent order of Hermite polynomials.
    p = jnp.floor(index / (Nn * Nm)).astype(int)
    m = jnp.floor((index - p * Nn * Nm) / Nn).astype(int)
    n = (index - p * Nn * Nm - m * Nn).astype(int)
    
    C_expanded = C[:, index, :, :, :, jnp.newaxis, jnp.newaxis, jnp.newaxis]
    Herm_x_expanded = Hermite(n, xi_x)[jnp.newaxis, jnp.newaxis, jnp.newaxis, jnp.newaxis, :, :, :]
    Herm_y_expanded = Hermite(m, xi_y)[jnp.newaxis, jnp.newaxis, jnp.newaxis, jnp.newaxis, :, :, :]
    Herm_z_expanded = Hermite(p, xi_z)[jnp.newaxis, jnp.newaxis, jnp.newaxis, jnp.newaxis, :, :, :]
    exp_expanded = jnp.exp(-(xi_x**2 + xi_y**2 + xi_z**2))[jnp.newaxis, jnp.newaxis, jnp.newaxis, jnp.newaxis, :, :, :]
    
    # Generate element of AW Hermite basis in 3D space.
    Hermite_term = C_expanded * (Herm_x_expanded * Herm_y_expanded * Herm_z_expanded * exp_expanded / 
                (jnp.sqrt(jnp.pi**3) * 2**((n + m + p) / 2) * jnp.sqrt(factorial(n)) * jnp.sqrt(factorial(m)) * jnp.sqrt(factorial(p))))
    
    return Hermite_term

# def rescale_C(C,Nn,Nm,index):
    
#     p = jnp.floor(index / (Nn * Nm)).astype(int)
#     m = jnp.floor((index - p * Nn * Nm) / Nn).astype(int)
#     n = (index - p * Nn * Nm - m * Nn).astype(int)
#     # debug.print('{},{},{},{},{},{}', p, m, n, factorial(n) * factorial(m) * factorial(p), jnp.abs(jnp.power(2,(n + m + p)/1)), index)
#     return C[:,index,...] / (2**((n + m + p)/2) * jnp.sqrt(factorial(n)) * jnp.sqrt(factorial(m)) * jnp.sqrt(factorial(p)))
    

def inverse_HF_transform(Ck, Lx, Ly, Lz, Nn, Nm, Np, xi_x, xi_y, xi_z):
    
    C = ifftn(ifftshift(Ck, axes=(-3, -2, -1)), axes=(-3, -2, -1))
    
    Nt, Nx, Ny, Nz = Ck.shape[0], Ck.shape[-2], Ck.shape[-3], Ck.shape[-1]
    Nvx, Nvy, Nvz = xi_x.shape[1], xi_x.shape[0], xi_x.shape[2]
    
    f = jnp.zeros((Nt, Ny, Nx, Nz, Nvy, Nvx, Nvz))
    for index in jnp.arange(Nn * Nm * Np):  
        f = f + generate_Hermite_term(xi_x, xi_y, xi_z, C.real, Nn, Nm, index)
    
    return f


import os
number_of_processors_to_use = 5 # Parallelization, this should divide total resolution
os.environ["XLA_FLAGS"] = f'--xla_force_host_platform_device_count={number_of_processors_to_use}'
from time import time
from jax import block_until_ready
import jax.numpy as jnp
from spectrax import simulation, load_parameters, plot, initialize_xv

# Read from input.toml
input_parameters, solver_parameters = load_parameters('input_2D_orszag_tang.toml')

print('Setting up simulation parameters...')
start_time = time()
deltaB = 0.2 # In-plane magnetic field amplitude. 
U0 = deltaB * input_parameters["Omega_cs"][0] / jnp.sqrt(input_parameters["mi_me"])
kx = 2*jnp.pi/input_parameters["Lx"]
ky = 2*jnp.pi/input_parameters["Ly"]
Nv = 201 # Number of velocity points in each dimension.
nvxyz = 100 # Number of velocity points in each dimension for the velocity grid.
max_min_v_factor = 5 # Factor for the maximum and minimum velocity in each dimension.

# Electron and ion fluid velocities.
Ue = lambda x, y, z: U0 * jnp.array([-jnp.sin(ky * y), jnp.sin(kx * x), -deltaB * input_parameters["Omega_cs"][0] * (2 * kx * jnp.cos(2 * kx * x) + ky * jnp.cos(ky * y))])
Ui = lambda x, y, z: U0 * jnp.array([-jnp.sin(ky * y), jnp.sin(kx * x), jnp.zeros_like(x)])

# Magnetic and electric fields.
B = lambda x, y, z: jnp.array([-deltaB * jnp.sin(ky * y), deltaB * jnp.sin(2 * kx * x), jnp.ones_like(x)])
E = lambda x, y, z: jnp.array([jnp.zeros_like(x), jnp.zeros_like(x), jnp.zeros_like(x)])

# Electron and ion distribution functions.
fe = (lambda x, y, z, vx, vy, vz: (1 / (((2 * jnp.pi) ** (3 / 2)) * input_parameters["vte"] ** 3) * 
                                jnp.exp(-((vx - Ue(x, y, z)[0])**2 + (vy - Ue(x, y, z)[1])**2 + (vz - Ue(x, y, z)[2])**2) / (2 * input_parameters["vte"] ** 2))))
fi = (lambda x, y, z, vx, vy, vz: (1 / (((2 * jnp.pi) ** (3 / 2)) * input_parameters["vti"] ** 3) * 
                                jnp.exp(-((vx - Ui(x, y, z)[0])**2 + (vy - Ui(x, y, z)[1])**2 + (vz - Ui(x, y, z)[2])**2) / (2 * input_parameters["vti"] ** 2))))

input_parameters["Ck_0"], input_parameters["Fk_0"] =  block_until_ready(
    initialize_xv(B, E, fe, fi, Nv, nvxyz, max_min_v_factor, input_parameters, **solver_parameters))
print(f"took {time() - start_time} seconds to initialize the simulation parameters.")

print('Starting simulation...')
start_time = time()
output = block_until_ready(simulation(input_parameters, **solver_parameters))
print(f"Runtime: {time() - start_time} seconds")

print('Plotting results...')
plot(output)

# print('Saving results...')
# jnp.savez('output_orszag.npz', **output)

# print("Loading results...")
# output = jnp.load('output_orszag.npz')


import sys

sys.path.append(r'/Users/csvega/Desktop/Madison/Code/Simulations')
sys.path.append(r'/Users/csvega/Desktop/Madison/Code/Vlasov-Maxwell_Spectral_Solver/Vlasov-MaxwellSpectralSolver')

import json
import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from JAX_VM_solver import VM_simulation
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
import time

   
# Simulation parameters.
Nx, Ny, Nz = 3, 1, 1
Lx, Ly, Lz = 1.0, 1.0, 1.0
Nn, Nm, Np, Ns = 80, 1, 1, 2 
mi_me, Ti_Te = 1836.0, 1.0
Omega_cs = jnp.array([1.0, 1.0 / mi_me])
qs = jnp.array([-1, 1])
alpha_e = jnp.array([0.1, 0.1, 0.1])
alpha_s = jnp.concatenate([alpha_e, alpha_e * jnp.sqrt(Ti_Te / mi_me)])
u_s = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
nu = 1.0
t_steps, t_max, dt = 401, 40, 0.01

# Initial conditions

kx = 2 * jnp.pi / Lx # Wavenumber.
dn = 0.0001 # Density fluctuation.

# Fourier components of magnetic and electric fields.
Fk_0 = jnp.zeros((6, 1, 3, 1), dtype=jnp.complex128)
Fk_0 = Fk_0.at[0, 0, 0, 0].set(dn / (2 * kx * Omega_cs[0]))
Fk_0 = Fk_0.at[0, 0, 2, 0].set(dn / (2 * kx * Omega_cs[0]))
    
# Hermite-Fourier components of electron and ion distribution functions.
Ce0_mk, Ce0_0, Ce0_k = 0 + 1j * (1 / (2 * alpha_s[0] ** 3)) * dn, 1 / (alpha_s[0] ** 3) + 0 * 1j, 0 - 1j * (1 / (2 * alpha_s[0] ** 3)) * dn
Ci0_0 = 1 / (alpha_s[3] ** 3) + 0 * 1j
Ck_0 = jnp.zeros((2 * Nn, 1, 3, 1), dtype=jnp.complex128)
Ck_0 = Ck_0.at[0, 0, 0, 0].set(Ce0_mk)
Ck_0 = Ck_0.at[0, 0, 1, 0].set(Ce0_0)
Ck_0 = Ck_0.at[0, 0, 2, 0].set(Ce0_k)
Ck_0 = Ck_0.at[Nn, 0, 1, 0].set(Ci0_0)

# # Save parameters into txt.
# with open('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\S13_Two_stream_instability_1D_HF.txt', 'w') as file:
#     file.write(f"Nx, Ny, Nz: {Nx}, {Ny}, {Nz}\n")
#     file.write(f"Nvx, Nvy, Nvz: {Nvx}, {Nvy}, {Nvz}\n")
#     file.write(f"Lx, Ly, Lz: {Lx}, {Ly}, {Lz}\n")
#     file.write(f"Nn, Nm, Np, Ns: {Nn}, {Nm}, {Np}, {Ns}\n")
#     file.write(f"mi_me: {mi_me}\n")
#     file.write(f"Omega_cs: {Omega_cs.tolist()}\n")
#     file.write(f"qs: {qs.tolist()}\n")
#     file.write(f"alpha_s: {alpha_s.tolist()}\n")
#     file.write(f"u_s: {u_s.tolist()}\n")
#     file.write(f"nu: {nu}\n")
#     file.write(f"t_steps, t_max: {t_steps}, {t_max}\n")
#     file.write(f"dt: {dt}\n")

# Run simulation.
start_time = time.time()
Ck, Fk, t = VM_simulation(Ck_0, Fk_0, qs, nu, Omega_cs, alpha_s, mi_me, u_s, Lx, Ly, Lz, Nx, Ny, Nz, Nn, Nm, Np, Ns, t_max, t_steps, dt)
end_time = time.time()
t.block_until_ready()
print(f"Runtime: {end_time - start_time} seconds")

# # Save results.
# jnp.save('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\Ck', Ck)
# jnp.save('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\Fk', Fk)
# jnp.save('C:\Cristian\Postdoc\Madison\Code\Simulations\Two_stream_instability_1D_HF\S13\\time', t)

####################################################################################################################################################
# Data analysis.

lambda_D = jnp.sqrt(1 / (2 * (1 / alpha_s[0] ** 2 + 1 / (mi_me * alpha_s[3] ** 2)))) # Debye length.
k_norm = jnp.sqrt(2) * jnp.pi * alpha_s[0] / Lx # Perturbation wavenumber normalized to the inverse of the Debye length.

# Set n = 0, k = 0 mode to zero to get array with time evolution of perturbation.
dCk = Ck.at[:, 0, 0, 1, 0].set(0)
dCk = dCk.at[:, Nn * Nm * Np, 0, 1, 0].set(0)

peaks, _ = find_peaks(jnp.abs(dCk[:, 0, 0, 0, 0].imag))
p = jnp.polyfit(t[peaks], jnp.log(jnp.abs(dCk[:, 0, 0, 0, 0].imag[peaks])), 1)

# Compute energy.
electron_energy_Ck = (0.5 * alpha_s[0] * alpha_s[1] * alpha_s[2]) * ((0.5 * (alpha_s[0] ** 2 + alpha_s[1] ** 2 + alpha_s[2] ** 2) + 
                                            (u_s[0] ** 2 + u_s[1] ** 2 + u_s[2] ** 2)) * Ck[:, 0, 0, int((Nx-1)/2), 0] + 
                                    jnp.sqrt(2) * alpha_s[0] * u_s[0] * Ck[:, 1, 0, int((Nx-1)/2), 0] + 
                                    (1 / jnp.sqrt(2)) * (alpha_s[0] ** 2) * Ck[:, 2, 0, int((Nx-1)/2), 0])
        
ion_energy_Ck = (0.5 * alpha_s[3] * alpha_s[4] * alpha_s[5]) * ((0.5 * (alpha_s[3] ** 2 + alpha_s[4] ** 2 + alpha_s[5] ** 2) + 
                                            (u_s[3] ** 2 + u_s[4] ** 2 + u_s[5] ** 2)) * Ck[:, Nn, 0, int((Nx-1)/2), 0] + 
                                    jnp.sqrt(2) * alpha_s[3] * u_s[3] * Ck[:, Nn + 1, 0, int((Nx-1)/2), 0] + 
                                    (1 / jnp.sqrt(2)) * (alpha_s[3] ** 2) * Ck[:, Nn + 2, 0, int((Nx-1)/2), 0])
                                                     
electric_energy_Fk = 0.5 * jnp.sum(jnp.abs(Fk[:, 0, 0, :, 0]) ** 2, axis=-1) * Omega_cs[0] ** 2



fig1, axes1 = plt.subplots(2, 3, figsize=(18, 10))

# Energy plots in linear scale.

# Electron energy.
axes1[0,0].plot(t, electron_energy_Ck + ion_energy_Ck, label='Plasma energy', linestyle='-', color='red', linewidth=3.0)
axes1[0,0].set_ylabel(r'plasma energy', fontsize=16)
axes1[0,0].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes1[0,0].legend().set_draggable(True)


# Electric energy.
axes1[0,1].plot(t, electric_energy_Fk, label='electric energy', linestyle='-', color='red', linewidth=3.0)
axes1[0,1].set_ylabel(r'electric energy', fontsize=16)
axes1[0,1].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes1[0,1].legend().set_draggable(True)


# Total energy.
axes1[0,2].plot(t, electron_energy_Ck + ion_energy_Ck + electric_energy_Fk, label='total energy', linestyle='-', color='red', linewidth=3.0)
axes1[0,2].set_ylabel(r'total energy', fontsize=16)
axes1[0,2].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes1[0,2].legend().set_draggable(True)


# Energy plots in log-linear scale.

# Electron energy.
axes1[1,0].plot(t, jnp.log10(electron_energy_Ck + ion_energy_Ck), label='plasma energy', linestyle='-', color='red', linewidth=3.0)
axes1[1,0].set_ylabel(r'$\log_{10}(\text{plasma energy})$', fontsize=16)
axes1[1,0].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes1[1,0].legend().set_draggable(True)


# Electric energy.
axes1[1,1].plot(t, jnp.log10(electric_energy_Fk), label='electric energy', linestyle='-', color='red', linewidth=3.0)
axes1[1,1].set_ylabel(r'$log_{10}(\text{electric energy})$', fontsize=16)
axes1[1,1].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes1[1,1].legend().set_draggable(True)


# Total energy.
axes1[1,2].plot(t, jnp.log10(electron_energy_Ck + ion_energy_Ck + electric_energy_Fk), label='total energy', linestyle='-', color='red', linewidth=3.0)
axes1[1,2].set_ylabel(r'$log_{10}(\text{total energy})$', fontsize=16)
axes1[1,2].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes1[1,2].legend().set_draggable(True)

fig1.suptitle(rf'$kv_{{th,e}}/\omega_{{pe}} = {k_norm:.2}, \nu = {nu}, u_e = {u_s[0]}, \alpha_e = {alpha_s[0]:.3}, N_x = {Nx}, N_n = {Nn}, \delta n = 0.001$', fontsize=14)



fig2, axes2 = plt.subplots(1, 2, figsize=(18, 10))

# Plot electron density fluctuation vs t.
axes2[0].plot(t, jnp.log(jnp.abs(dCk[:, 0, 0, 0, 0].imag)), label='$log(|\delta C^e_{000,k}|)$', linestyle='-', color='red', linewidth=3.0)
axes2[0].plot(t, p[0] * t + p[1], label=f'damping rate = {p[0]:.3}', linestyle='-', color='black', linewidth=3.0)
axes2[0].set_ylabel(r'$log(|\delta C^e_{000,k}|)$', fontsize=16)
axes2[0].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes2[0].set_xlim(0.0, t_max)
axes2[0].legend().set_draggable(True)

# Plot ion density fluctuation vs t.
axes2[1].plot(t, jnp.log(jnp.abs(dCk[:, Nn, 0, 0, 0].imag)), label='$log(|\delta C^i_{000,k}|)$', linestyle='-', color='red', linewidth=3.0)
axes2[1].set_ylabel(r'$log(|\delta C^i_{000,k}|)$', fontsize=16)
axes2[1].set_xlabel(r'$t\omega_{pe}$', fontsize=16)
axes2[1].set_xlim(0.0, t_max)
axes2[1].legend().set_draggable(True)

fig2.suptitle(rf'$kv_{{th,e}}/\omega_{{pe}} = {k_norm:.2}, u_e/c = {u_s[0]}, \lambda_D/d_e = {lambda_D:.1e}, m_i/m_e = {mi_me}, N_n = {Nn}$', fontsize=14)

plt.show()